//Core bitsensor interface
var ApiConnector = require('./api_connector');
var Config = require('./config');
var Datapoint = require('./compiled').proto.Datapoint;
var RWLock = require('rwlock');
var lock = new RWLock();

/**
 * Entry point for starting the BitSensor Web Application Security plugin.
 *
 * @constructor
 */
var BitSensor = function () {

  /**
   * @type {Handler[]}
   */
  this.preHandlers = [];
  this.postHandlers = [];
  this.session = require("continuation-local-storage").createNamespace("bitsensor");
};

BitSensor.prototype.runSession = function (callback) {
  this.session.run(() => {
    if (!this.isInitialized()) {
      this.initialize();
    }
    this.traceIncrement();
    if (typeof callback === "function")
      callback();
    this.ship();
  });
};

BitSensor.prototype.initialize = function () {
  if (this.isInitialized()) {
    return;
  }

  this.setDatapoint(new Datapoint());
  this.initializeEnabled();
  this.runPreHandlers();
};

BitSensor.prototype.isInitialized = function () {
  return this.session && !!this.session.get('datapoint');
};

BitSensor.prototype.apply = function (consumer) {
  if (this.hasBeenDisabled()) {
    return;
  }

  if (!this.isInitialized()) {
    this.runSession(() => {
      consumer.call(this, this.getDatapoint());
    });
    return;
  }

  consumer.call(this, this.getDatapoint());
};


BitSensor.prototype.getDatapoint = function () {
  return bitSensor.session.get('datapoint');
};

BitSensor.prototype.setDatapoint = function (datapoint) {
  this.session.set("datapoint", datapoint);
};

BitSensor.prototype.ship = function () {
  if (this.hasBeenDisabled())
    return;

  if (this.traceDecrement() > 0) {
    return;
  }

  this.runPostHandlers();
  this.log();
  this.clear();
};

BitSensor.prototype.log = function () {
  if (!this._config) {
    return;
  }
  new ApiConnector().from(this._config.user, this._config.apiKey)
    .to(this._config.uri)
    .with(this.getDatapoint())
    .post(ApiConnector.ACTION_LOG)
    .send();
};

BitSensor.prototype.clear = function () {
  this.session.set('isDisabled', null);
  this.session.set('isHandlingRequest', null);
  this.session.set('datapoint', null);
};

BitSensor.prototype.hasBeenDisabled = function () {
  if (this.getDisabled() == null) {
    return false;
  }

  return this.getDisabled() > 0;
};

BitSensor.prototype.setDisabled = function () {
  this.initializeEnabled();
  this.session.set('isDisabled', this.getDisabled() + 1);

  return this.getDisabled();
};

BitSensor.prototype.setEnabled = function () {
  this.initializeEnabled();
  if (this.getDisabled() > 0) {
    this.session.set('isDisabled', this.getDisabled() - 1);
  }

  return this.getDisabled();
};

BitSensor.prototype.initializeEnabled = function () {
  if (this.getDisabled() == null) {
    this.session.set('isDisabled', 0);
  }
  if (this.getHandlingRequest() == null) {
    this.session.set('isHandlingRequest', 0);
  }
};

BitSensor.prototype.getDisabled = function () {
  return bitSensor.session.get('isDisabled');
}

BitSensor.prototype.getHandlingRequest = function () {
  return bitSensor.session.get('isHandlingRequest');
}

BitSensor.prototype.traceIncrement = function () {
  if (this.getHandlingRequest() == null) {
    this.initialize();
  }

  bitSensor.session.set('isHandlingRequest', bitSensor.getHandlingRequest() + 1);
};

BitSensor.prototype.traceDecrement = function () {
  if (this.getHandlingRequest() == null) {
    this.initialize();
  }
  if (bitSensor.getHandlingRequest() > 0) {
    bitSensor.session.set('isHandlingRequest', bitSensor.getHandlingRequest() - 1);
  }

  return this.getHandlingRequest();
};


BitSensor.prototype.addPreHandler = function (handler) {
  this.preHandlers.push(handler);
};

BitSensor.prototype.addPostHandler = function (handler) {
  this.postHandlers.push(handler);
};


BitSensor.prototype.runPreHandlers = function () {
  this.preHandlers.forEach((handler) => {
    handler.handle(this.getDatapoint());
  });
};

BitSensor.prototype.runPostHandlers = function () {
  this.postHandlers.forEach((handler) => {
    handler.handle(this.getDatapoint());
  });
};

BitSensor.prototype.config = function (config) {
  this._config = new Config(config);
}

module.exports = BitSensor;